#ifndef PSTEP_H
#define PSTEP_H

#include<stdio.h>
#include<stdlib.h>
#include"util.h"
#include<stdlib.h>

#ifndef ONEEDGE_DEFINED
#define ONEEDGE_DEFINED

  typedef struct oneedge {
          /* pointer to other three edges */
          struct oneedge *up;
          struct oneedge *downleft;
          struct oneedge *downright;
          int numedge;    /* number of edge */
          uli edgeinfo;   /* penalty of this edge */
          uli edgesum;    /* penalty sums of this edge */
          uli edgemat;    /* penalty submatr. of this edge */
          int *edgemap;   /* pointer to the local edgemap */
                          /* OR list of descendances */
  } ONEEDGE;

#endif /* ! ONEEDGE_DEFINED */


#ifndef FPRINTF
#  define FPRINTF fprintf
#  define STDOUT     stdout
#  define STDOUTFILE STDOUT,
#endif


/******************************************************************************/
/* functions for representing and building puzzling step trees                */
/******************************************************************************/

/* initialize tree with the following starting configuration

                 2
          0  +------- C(=2)
  A(=0) -----+
             +------- B(=1)
                 1
                                */
void newinittree(ONEEDGE **edge, 
              int     **edgeofleaf, 
              int       Maxspc, 
              int       Maxbrnch, 
              int      *nextedge, 
              int      *nextleaf);

/* add next leaf on the specified edge */
void newaddnextleaf(int      dockedge,
                 ONEEDGE *edge, 
                 int     *edgeofleaf, 
                 int      Maxspc, 
                 int      Maxbrnch, 
                 int     *nextedge, 
                 int     *nextleaf);

/* free memory (to be called after inittree) */
void newfreetree(ONEEDGE *edge, 
              int     *edgeofleaf, 
              int      Maxspc);

/* writes OTU sitting on edge ed */
void writeOTU(FILE    *outfp, 
              int      ed, 
              ONEEDGE *edge, 
              int     *edgeofleaf, 
              int      nextedge, 
              int      nextleaf,
              int     *column,
              int     *trueID);

/* write tree */
void writetree(FILE   *outfp,
              ONEEDGE *edge, 
              int     *edgeofleaf, 
              int      nextedge, 
              int      nextleaf,
              int     *trueID);

/* writes OTU sitting on edge ed with debug info */
void newwriteOTU(FILE    *outfp, 
              int      ed, 
              ONEEDGE *edge, 
              int     *edgeofleaf, 
              int      nextedge, 
              int      nextleaf,
              int     *column,
              int     *trueID);

/* write tree with debug info */
void newwritetree(FILE   *outfp,
              ONEEDGE *edge, 
              int     *edgeofleaf, 
              int      nextedge, 
              int      nextleaf,
              int     *trueID);

/* clear all edgeinfos */
void newresetedgeinfo(ONEEDGE *edge, 
                   int      nextedge);

/* ajust edge penalties recursively */
void newincrementedgeinfo(ONEEDGE  *edge,
                          int      *edgeofleaf,
                          ulimatrix mat,
                          ulivector vec,
                          int      *trueID);

/* checks which edge has the lowest edgeinfo
   if there are several edges with the same lowest edgeinfo,
   one of them will be selected randomly */
void newminimumedgeinfo(ONEEDGE *edge, 
                     int     *edgeofleaf, 
                     int      nextedge, 
                     int      nextleaf,
                     int     *minedge, 
                     uli     *mininfo, 
                     int     *howmany);


#endif /* PSTEP_H */
