/*
 * pustep.c
 *
 *
 * Part of TREE-PUZZLE 5.0 (June 2000)
 *
 * (c) 1999-2000 by Heiko A. Schmidt, Korbinian Strimmer,
 *                  M. Vingron, and Arndt von Haeseler
 * (c) 1995-1999 by Korbinian Strimmer and Arndt von Haeseler
 *
 * All parts of the source except where indicated are distributed under
 * the GNU public licence.  See http://www.opensource.org for details.
 */


#define EXTERN
#define FILENAMELEN 256

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <ctype.h>
#include <string.h>
#include <limits.h>
#include <float.h>
#include "util.h"
#include "newpstep.h"
#include "oldpstep.h"

#ifndef FPRINTF
#  define FPRINTF    fprintf
#  define STDOUT     stdout
#  define STDOUTFILE STDOUT,
#endif

int deb = 0;

unsigned char readquartet(int a, int b, int c, int d, 
                          unsigned char *quartetinfo);
void writequartet(int a, int b, int c, int d, unsigned char info,
                  unsigned char *quartetinfo);
void readquartetlh(int a, int b, int c, int d, 
                   double *d1, double *d2, double *d3, 
                   double *lhinfo);
void writequartetlh(int a, int b, int c, int d, 
                    double d1, double d2, double d3,
                    double *lhinfo);
unsigned char weightquart(double d1, double d2, double d3);



char **spcnames;

int fputid(FILE *fp, int n)
{
   fprintf(fp, "%s", spcnames[n]);
   return(strlen(spcnames[n]));
}

void fputid10(FILE *fp, int n)
{
   fprintf(fp, "%-10s", spcnames[n]);
}

char topostr[8][4] = {"000", "001", "010", "011",
                      "100", "101", "110", "111"};

#ifndef _UTIL_
typedef unsigned long int uli;
#endif

typedef struct {
  char           *fname;
  char            format[2];
  FILE           *fstream;
  int             nspec;
  int            *translat;
  char          **names;
  unsigned char  *quartetinfo;
  double         *lhinfo;
} qblocktype;

/*********************************************************************
*  read/write array with all quartets                                *
*********************************************************************/


void sort4ints(int *a, int *b, int *c, int *d)
{
  int t;
  if (*a > *b) { t=*a; *a=*b; *b=t; }
  if (*a > *c) { t=*a; *a=*c; *c=t; }
  if (*a > *d) { t=*a; *a=*d; *d=t; }
  
  if (*b > *c) { t=*b; *b=*c; *c=t; }
  if (*b > *d) { t=*b; *b=*d; *d=t; }
  
  if (*c > *d) { t=*c; *c=*d; *d=t; }
}

/******************************************************************************/

void sort3doubles(double num[3], int order[3])
{
        if (num[0] > num[1]) {
                if(num[2] > num[0]) {
                        order[0] = 2;
                        order[1] = 0;
                        order[2] = 1;
                } else if (num[2] < num[1]) {
                        order[0] = 0;
                        order[1] = 1;
                        order[2] = 2;
                } else {
                        order[0] = 0;
                        order[1] = 2;
                        order[2] = 1;
                }
        } else {
                if(num[2] > num[1]) {
                        order[0] = 2;
                        order[1] = 1;
                        order[2] = 0;
                } else if (num[2] < num[0]) {
                        order[0] = 1;
                        order[1] = 0;
                        order[2] = 2;
                } else {
                        order[0] = 1;
                        order[1] = 2;
                        order[2] = 0;
                }
        }
} 


/******************************************************************************/

void num2quart(uli qnum, int *a, int *b, int *c, int *d)
{
        double temp;
        uli aa, bb, cc, dd;
        uli lowval=0, highval=0;

        aa=0; bb=1; cc=2; dd=3;

        temp = (double)(24 * qnum);
        temp = sqrt(temp);
        temp = sqrt(temp);
        /* temp = pow(temp, (double)(1/4)); */
        dd = (uli) floor(temp) + 1;
        if (dd < 3) dd = 3;
        lowval =  (uli) dd*(dd-1)*(dd-2)*(dd-3)/24;
        highval = (uli) (dd+1)*dd*(dd-1)*(dd-2)/24;
        if (lowval >= qnum)
            while ((lowval > qnum)) {
                dd -= 1; lowval = (uli) dd*(dd-1)*(dd-2)*(dd-3)/24;
            }
        else {
            while (highval <= qnum) {
                dd += 1; highval = (uli) (dd+1)*dd*(dd-1)*(dd-2)/24;
            }
            lowval = (uli) dd*(dd-1)*(dd-2)*(dd-3)/24;
        }
        qnum -= lowval;
        if (qnum > 0) {
            temp = (double)(6 * qnum);
            temp = pow(temp, (double)(1/3));
            cc = (uli) floor(temp);
            if (cc < 2) cc= 2;
            lowval =  (uli) cc*(cc-1)*(cc-2)/6;
            highval = (uli) (cc+1)*cc*(cc-1)/6;
            if (lowval >= qnum)
                while ((lowval > qnum)) {
                   cc -= 1; lowval = (uli) cc*(cc-1)*(cc-2)/6;
                }
            else {
                while (highval <= qnum) {
                   cc += 1; highval = (uli) (cc+1)*cc*(cc-1)/6;
                }
                lowval = (uli) cc*(cc-1)*(cc-2)/6;
            }
            qnum -= lowval;
            if (qnum > 0) {
                temp = (double)(2 * qnum);
                temp = sqrt(temp);
                bb = (uli) floor(temp);
                if (bb < 1) bb= 1;
                lowval =  (uli) bb*(bb-1)/2;
                highval = (uli) (bb+1)*bb/2;
                if (lowval >= qnum)
                    while ((lowval > qnum)) {
                        bb -= 1; lowval = (uli) bb*(bb-1)/2;
                    }
                else {
                    while (highval <= qnum) {
                       bb += 1; highval = (uli) (bb+1)*bb/2;
                    }
                    lowval = (uli) bb*(bb-1)/2;
                }
                qnum -= lowval;
                if (qnum > 0) {
                   aa = (uli) qnum;
                   if (aa < 0) aa= 0;
                }
            }
        }
        *d = (int)dd;
        *c = (int)cc;
        *b = (int)bb;
        *a = (int)aa;
}  /* num2quart */

/******************/

uli numquarts(int maxspc)
{
   uli tmp;
   int a, b, c, d;

   if (maxspc < 4)
     return (uli)0;
   else {
      maxspc--;
      a = maxspc-3;
      b = maxspc-2;
      c = maxspc-1;
      d = maxspc;

      tmp = (uli) 1 + a +
            (uli) b * (b-1) / 2 +
            (uli) c * (c-1) * (c-2) / 6 +
            (uli) d * (d-1) * (d-2) * (d-3) / 24;
      return (tmp);
   }
}  /* numquarts */

/******************/

uli quartnum2bytes(int maxquart)
{
  /* compute number of bytes */
  if (maxquart % 2 == 0) { /* even number */
     return((maxquart)/2);
  } else { /* odd number */
     return((maxquart + 1)/2);
  }

}

/******************/

uli specnum2bytes(int maxspc)
{
  return (quartnum2bytes(numquarts(maxspc)));
}

/******************/

uli quart2num (int a, int b, int c, int d)
{
      uli tmp;
      if ((a>b) || (b>c) || (c>d)) {
         fprintf(stderr, "Warning PP5: not (%d<%d<%d<%d) !!!\n",a,b,c,d);
         sort4ints(&a, &b, &c, &d);
      }
      tmp = (uli) a +
            (uli) b * (b-1) / 2 +
            (uli) c * (c-1) * (c-2) / 6 +
            (uli) d * (d-1) * (d-2) * (d-3) / 24;
      return (tmp);
}  /* quart2num */

/******************/

void printallquarts(	FILE       *fp, 
			qblocktype  qb, 
			int         idx, 
			int         brackets, 
			int         topol, 
			int         bits, 
			int         names,
			int         formatflag);


void printverboseallquarts(FILE *fp, qblocktype qb)
{
  printallquarts(	fp, 
			qb,
			1,  /* idx */ 
			1,  /* brackets */ 
			1,  /* topol */ 
			1,  /* bits */ 
			1,  /* names */
			0   /* 0-old, 1-BB, 2-BA, 3-LB, 4-LA */);
}


void printasciiallquarts(FILE *fp, qblocktype qb)
{
  printallquarts(	fp, 
			qb,
			0,  /* idx */ 
			1,  /* brackets */ 
			0,  /* topol */ 
			1,  /* bits */ 
			0,  /* names */
			2   /* 0-old, 1-BB, 2-BA, 3-LB, 4-LA */);
}


/******************/

/*  flag=0	old allquart binary  */
/*  flag=1	allquart binary  */
/*  flag=2	allquart ACSII   */
/*  flag=3	quartlh  binary  */
/*  flag=4	quartlh  ASCII   */

void writetpqfheader(int            nspec,
                     char         **names,
                     FILE          *ofp,
                     int            flag)
{ int            currspec;

  if (flag == 0) {
     unsigned long  nquart;
     unsigned long  blocklen;

     nquart = numquarts(nspec);
     /* compute number of bytes */
     if (nquart % 2 == 0) { /* even number */
        blocklen = (nquart)/2;
     } else { /* odd number */
        blocklen = (nquart + 1)/2;
     }
     /* FPRINTF(STDOUTFILE "Writing quartet file: %s\n", filename); */
     fprintf(ofp, "TREE-PUZZLE\n5.0\n\n");
     fprintf(ofp, "species: %d\n",       nspec);
     fprintf(ofp, "quartets: %lu\n",     nquart);
     fprintf(ofp, "bytes: %lu\n\n",      blocklen);


     /* fwrite(&(quartetinfo[0]), sizeof(char), blocklen, ofp); */
  }

  if (flag == 1) fprintf(ofp, "##TPQF-BB (CHECKQUART)\n%d\n", nspec);
  if (flag == 2) fprintf(ofp, "##TPQF-BA (CHECKQUART)\n%d\n", nspec);
  if (flag == 3) fprintf(ofp, "##TPQF-LB (CHECKQUART)\n%d\n", nspec);
  if (flag == 4) fprintf(ofp, "##TPQF-LA (CHECKQUART)\n%d\n", nspec);

  for (currspec=0; currspec<nspec; currspec++) {
     fprintf(ofp, "%s\n", names[currspec]);
  } /* for each species */
  fprintf(ofp, "\n");

} /* writetpqfheader */



/******************/

/* check the branching structure between the leaves (not the taxa!)
   A, B, C, and I (A, B, C, I don't need to be ordered). As a result,
   the two leaves that are closer related to each other than to leaf I
   are found in chooseA and chooseB. If the branching structure is
   not uniquely defined, ChooseA and ChooseB are chosen randomly
   from the possible taxa */
void checkquartet(int A, int B, int C, int I, 
                  int *trueID, unsigned char *quartetinfo,
                  int *chooseA, int *chooseB)
{
	int i, j, a, b, taxon[5], leaf[5], ipos;
	unsigned char qresult;
	int notunique = FALSE;

	/* The relationship between leaves and taxa is defined by trueID */
	taxon[1] = trueID[A]; /* taxon number */
	leaf[1] = A;          /* leaf number  */
	taxon[2] = trueID[B];
	leaf[2] = B;
	taxon[3] = trueID[C];
	leaf[3] = C;
	taxon[4] = trueID[I];
	leaf[4] = I;

	/* sort for taxa */
	/* Source: Numerical Recipes (PIKSR2.C) */
	for (j = 2; j <= 4; j++) {
		a = taxon[j];
		b = leaf[j];
		i = j-1;
		while (i > 0 && taxon[i] > a) {
			taxon[i+1] = taxon[i];
			leaf[i+1] = leaf[i];
			i--;
		}
		taxon[i+1] = a;
		leaf[i+1] = b;
	}

	/* where is leaf I ? */
	ipos = 1;
	while (leaf[ipos] != I) ipos++;

	/* look at sequence quartet */
        qresult = readquartet(taxon[1], taxon[2], taxon[3], taxon[4], 
                          quartetinfo);

	/* chooseA and chooseB */
	do {	
		switch (qresult) {
		
			/* one single branching structure */
		
			/* 001 */
			case 1:		if (ipos == 1 || ipos == 2) {
							*chooseA = leaf[3];
							*chooseB = leaf[4];
						} else {
							*chooseA = leaf[1];
							*chooseB = leaf[2];
						}
						notunique = FALSE;
						break;

			/* 010 */
			case 2:		if (ipos == 1 || ipos == 3) {
							*chooseA = leaf[2];
							*chooseB = leaf[4];
						} else {
							*chooseA = leaf[1];
							*chooseB = leaf[3];
						}
						notunique = FALSE;
						break;

			/* 100 */
			case 4:		if (ipos == 1 || ipos == 4) {
							*chooseA = leaf[2];
							*chooseB = leaf[3];
						} else {
							*chooseA = leaf[1];
							*chooseB = leaf[4];
						}
						notunique = FALSE;
						break;

			/* two possible branching structures */		

			/* 011 */
			case 3:		if (randominteger(2)) qresult = 1;
						else qresult = 2;
fprintf(stderr, "e011\n");
						notunique = TRUE;
						break;

			/* 101 */
			case 5:		if (randominteger(2)) qresult = 1;
						else qresult = 4;
fprintf(stderr, "e101\n");
						notunique = TRUE;
						break;

			/* 110 */
			case 6:		if (randominteger(2)) qresult = 2;
						else qresult = 4;
fprintf(stderr, "e110\n");
						notunique = TRUE;
						break;

			/* three possible branching structures */

			/* 111 */
			case 7:		qresult = (1 << randominteger(3)); /* 1, 2, or 4 */
						notunique = TRUE;
fprintf(stderr, "e111\n");
						break;

			default:	/* Program error [checkquartet] */
#if 0
#if 				PARALLEL
					FPRINTF(STDOUTFILE "\n\n\n(%2d)HALT: PLEASE REPORT ERROR K-PARALLEL TO DEVELOPERS (%d,%d,%d,%d) = %ld\n\n\n", 
					PP_Myid, taxon[1], taxon[2], taxon[3], taxon[4],
					quart2num(taxon[1], taxon[2], taxon[3], taxon[4]));
#				else /* not PARALLEL */
					FPRINTF(STDOUTFILE "\n\n\nHALT: PLEASE REPORT ERROR K TO DEVELOPERS\n\n\n");
#				endif /* not PARALLEL */
#endif
				*chooseA = 0;
				*chooseB = 0;
				break;
						
		}
	} while (notunique);

	return;
}


/******************/

void printallquarts(	FILE       *fp, 
			qblocktype  qb, 
			int         idx, 
			int         brackets, 
			int         topol, 
			int         bits, 
			int         names,
			int         formatflag)
{
  int a, b, c, d;
  uli qnum, qbytes;
  unsigned topos;

  qnum = numquarts(qb.nspec);
  qbytes = quartnum2bytes(qnum);

  writetpqfheader(qb.nspec, qb.names, fp, formatflag);

  if (qb.quartetinfo == NULL)
    fprintf(fp, "NO QUARTETBLOCK READ!\n");
  else
    for (d = 3; d < qb.nspec; d++)
      for (c = 2; c < d; c++)
        for (b = 1; b < c; b++)
          for (a = 0; a < b; a++) {
	    topos = readquartet(a, b, c, d, qb.quartetinfo);
	    if (idx)
		fprintf(fp, "%lu\t", quart2num(a, b, c, d));
	    if (brackets)
		fprintf(fp, "(%d,%d,%d,%d)\t", a, b, c, d);
	    if (topol)
		fprintf(fp, "%d\t", topos);
	    if (bits)
		fprintf(fp, "%s\t", topostr[topos]);
	    if (names)
		fprintf(fp, "%s,%s,%s,%s", qb.names[a], qb.names[b],
                                           qb.names[c], qb.names[d]);
  		fprintf(fp, "\n"); fflush(stdout);
          }
}

/************************/

void printbinlh(qblocktype *qb)
{
  if (qb->fstream == NULL) {
    qb->fstream = fopen(qb->fname, "w");
    if (qb->fstream == NULL) {
      fprintf( stderr, "\nCan't open file %s to write!\n\n", qb->fname);
      exit (1);
    }
  }

  writetpqfheader(qb->nspec, qb->names, qb->fstream, 3);

  if (qb->lhinfo == NULL)
    fprintf(stderr, "NO QUARTETBLOCK READ!\n");
  else
    fwrite(qb->lhinfo, sizeof(double), 3 * numquarts(qb->nspec), qb->fstream);

  fclose(qb->fstream);
  qb->fstream = NULL;
}

/************************/

void printbinallquarts(qblocktype *qb)
{
  uli qnum, qbytes;
  int            currspec;

  if (qb->fstream == NULL) {
    qb->fstream = fopen(qb->fname, "w");
    if (qb->fstream == NULL) {
      fprintf( stderr, "\nCan't open file %s to write!\n\n", qb->fname);
      exit (1);
    }
  }

  fprintf(qb->fstream, "TREE-PUZZLE\n");
  fprintf(qb->fstream, "5.0\n");

  qnum = numquarts(qb->nspec);
  qbytes = quartnum2bytes(qnum);
  fprintf(qb->fstream, "\n");
  fprintf(qb->fstream, "species: %d\n", qb->nspec);
  fprintf(qb->fstream, "quartets: %lu\n", qnum);
  fprintf(qb->fstream, "bytes: %lu\n", qbytes);
  fprintf(qb->fstream, "\n");

  for (currspec=0; currspec < (qb->nspec); currspec++) 
     fprintf(qb->fstream, "%s\n", (qb->names)[currspec]);
  fprintf(qb->fstream, "\n");

  if (qb->quartetinfo == NULL)
    fprintf(stderr, "NO QUARTETBLOCK READ!\n");
  else
    fwrite(qb->quartetinfo, sizeof(char), qbytes, qb->fstream);

  fclose(qb->fstream);
  qb->fstream = NULL;
}

/************************/

void newreadheader(qblocktype *qb, int closeflag)
{ 
  int           currspec;
  char          dummyprog[128];
  char		tmpchr;
  int           strp;
  int           flag;


  if (qb->fstream != NULL) {
    fclose(qb->fstream);
    qb->fstream = NULL;
  }

  qb->fstream = fopen(qb->fname, "r");
  if (qb->fstream == NULL) {
    fprintf( stderr, "\nCan't open file %s\n\n", qb->fname);
    exit (1);
  }

  fscanf(qb->fstream, "##TPQF-%c%c\n", &(qb->format[0]), &(qb->format[1]));
  flag = 0;
  strp = 0;
  tmpchr = fgetc(qb->fstream);
  while (tmpchr != '\n') {
     if (!flag && (tmpchr == '('))
          flag = 1;
     else if (tmpchr == ')')
               flag = 0;
          else if (flag)
                  dummyprog[strp++] = tmpchr;
     tmpchr = fgetc(qb->fstream);
  }
  dummyprog[strp] = '\0';
  fscanf(qb->fstream, "%d\n",   &(qb->nspec));

  fprintf(stdout, "Reading quartet file: %s\n", qb->fname);
  fprintf(stdout, "   type: %c%c\n", qb->format[0], qb->format[1]);
  fprintf(stdout, "   generated by: %s\n", dummyprog);

  fprintf(stdout, "   %d species\n", qb->nspec);

  qb->names       = malloc(qb->nspec * sizeof(char *));

  for (currspec=0; currspec<qb->nspec; currspec++) {

     (qb->names)[currspec] = (char *)malloc(11 * sizeof(char));
     fscanf(qb->fstream, "%s\n", (qb->names)[currspec]);
     fprintf(stdout, "   %3d. %s\n", currspec+1, (qb->names)[currspec]);

  } /* for each species */
  fprintf(stdout, "\n");

  fscanf(qb->fstream, "\n");

  if (closeflag) {
     fclose(qb->fstream);
     qb->fstream = NULL;
  }

} /* readallquart */

/************************/

void newreadallquarts(qblocktype    *qb)
{ unsigned long nquart;
  unsigned long blocklen;
  int           a,b,c,d;
#ifdef VERBOSE
  double        d1,d2,d3;
#endif
  float         l1,l2,l3;
  unsigned long n;

  newreadheader(qb, 0);

  nquart = numquarts(qb->nspec);

  if ((qb->format[0] == 'B') && (qb->format[1] == 'B')) {   /* topology bits in binary format */
      fprintf(stderr, "%c%c\n",qb->format[0],qb->format[1]);
      qb->quartetinfo = malloc(specnum2bytes(qb->nspec) * sizeof(unsigned char));
      fread(&(qb->quartetinfo[0]), sizeof(unsigned char), blocklen, qb->fstream);
  }

  if ((qb->format[0] == 'L') && (qb->format[1] == 'B')) {   /* log-LH in binary format */
      fprintf(stderr, "%c%c\n",qb->format[0],qb->format[1]);
      qb->lhinfo = malloc(3 * numquarts(qb->nspec) * sizeof(double));
      fread(&(qb->lhinfo[0]), sizeof(double), 3*numquarts(qb->nspec), qb->fstream);
#ifdef VERBOSE
      for (n=0; n<nquart; n++){
         num2quart(n, &a, &b, &c, &d);
         readquartetlh(a, b, c, d, &d1, &d2, &d3, qb->lhinfo);
         printf("(%d,%d,%d,%d)\t%f\t%f\t%f\n",a,b,c,d, d1, d2, d3);
      }
#endif
  }

  if ((qb->format[0] == 'L') && (qb->format[1] == 'A')) {   /* log-LH in ASCII format */
      fprintf(stderr, "%c%c\n",qb->format[0],qb->format[1]);
      qb->lhinfo = malloc(3 * numquarts(qb->nspec) * sizeof(double));
      /* fread(&(qb->hlinfo[0]), sizeof(unsigned char), blocklen, qb->fstream); */
      for (n=0; n<nquart; n++){
         fscanf(qb->fstream, "(%d,%d,%d,%d)\t%f\t%f\t%f\n",&a,&b,&c,&d, &l1, &l2, &l3);
#ifdef VERBOSE
         printf("(%d,%d,%d,%d)\t%f\t%f\t%f\n",a,b,c,d, l1, l2, l3);
#endif
         writequartetlh(a, b, c, d, (double)l1, (double)l2, (double)l3, qb->lhinfo);
      }
  }

  fclose(qb->fstream);
  qb->fstream = NULL;

} /* readallquart */

/************************/

void freshreadheader(qblocktype    *qb)
{ unsigned long nquart;
  unsigned long blocklen;
  int           currspec;
  char          dummyversion[128];
  char          dummyprog[128];


  if (qb->fstream == NULL) {
    qb->fstream = fopen(qb->fname, "r");
    if (qb->fstream == NULL) {
      fprintf( stderr, "\nCan't open file %s\n\n", qb->fname);
      exit (1);
    }
  }
  fscanf(qb->fstream, "%s\n", dummyprog);
  fscanf(qb->fstream, "%s\n\n", dummyversion);
  
  fscanf(qb->fstream, "species: %d\n",   &(qb->nspec));
  fscanf(qb->fstream, "quartets: %lu\n", &nquart);
  fscanf(qb->fstream, "bytes: %lu\n\n",  &blocklen);

  fprintf(stdout, "Reading quartet file: %s\n", qb->fname);
  fprintf(stdout, "   generated by: %s %s\n", dummyprog, dummyversion);

  fprintf(stdout, "   %d species, %lu quartets, %lu bytes\n", 
                   qb->nspec, nquart, blocklen);

  qb->names       = malloc(qb->nspec * sizeof(char *));

  for (currspec=0; currspec<qb->nspec; currspec++) {

     (qb->names)[currspec] = (char *)malloc(11 * sizeof(char));
     fscanf(qb->fstream, "%s\n", (qb->names)[currspec]);
     fprintf(stdout, "   %3d. %s\n", currspec+1, (qb->names)[currspec]);

  } /* for each species */
  fprintf(stdout, "\n");

/*
  qb->quartetinfo = malloc(blocklen * sizeof(unsigned char));
  fread(&(qb->quartetinfo[0]), sizeof(unsigned char), blocklen, qb->fstream); 
*/
  fclose(qb->fstream);
  qb->fstream = NULL;

} /* readallquart */

/************************/

void freshreadallquarts(qblocktype    *qb)
{ unsigned long nquart;
  unsigned long blocklen;
  int           currspec;
  char          dummyversion[128];
  char          dummyprog[128];


  if (qb->fstream == NULL) {
    qb->fstream = fopen(qb->fname, "r");
    if (qb->fstream == NULL) {
      fprintf( stderr, "\nCan't open file %s\n\n", qb->fname);
      exit (1);
    }
  }
  fscanf(qb->fstream, "%s\n", dummyprog);
  fscanf(qb->fstream, "%s\n\n", dummyversion);
  
  fscanf(qb->fstream, "species: %d\n",   &(qb->nspec));
  fscanf(qb->fstream, "quartets: %lu\n", &nquart);
  fscanf(qb->fstream, "bytes: %lu\n\n",  &blocklen);

  fprintf(stdout, "Reading quartet file: %s\n", qb->fname);
  fprintf(stdout, "   generated by: %s %s\n", dummyprog, dummyversion);

  fprintf(stdout, "   %d species, %lu quartets, %lu bytes\n", 
                   qb->nspec, nquart, blocklen);

  qb->quartetinfo = malloc(blocklen * sizeof(unsigned char));
  qb->names       = malloc(qb->nspec * sizeof(char *));

  for (currspec=0; currspec<qb->nspec; currspec++) {

     (qb->names)[currspec] = (char *)malloc(11 * sizeof(char));
     fscanf(qb->fstream, "%s\n", (qb->names)[currspec]);
     fprintf(stdout, "   %3d. %s\n", currspec+1, (qb->names)[currspec]);

  } /* for each species */
  fprintf(stdout, "\n");

  fread(&(qb->quartetinfo[0]), sizeof(unsigned char), blocklen, qb->fstream);
  fclose(qb->fstream);
  qb->fstream = NULL;

} /* freshreadallquart */

/************************/


void readquartetlh(int a, int b, int c, int d, 
                   double *d1, double *d2, double *d3, double *lhinfo)
{
        uli qnum;

	sort4ints(&a,&b,&c,&d);
        qnum = (uli) a
                        + (uli) b*(b-1)/2
                        + (uli) c*(c-1)*(c-2)/6
                        + (uli) d*(d-1)*(d-2)*(d-3)/24;
        qnum *= 3;
        *d1 = lhinfo[qnum];
        *d2 = lhinfo[qnum + 1];
        *d3 = lhinfo[qnum + 2];
} /* readquartetlh */

/************************/

void writequartetlh(int a, int b, int c, int d, 
                    double d1, double d2, double d3,
                    double *lhinfo)
{
        uli qnum;

        qnum = (uli) a
                        + (uli) b*(b-1)/2
                        + (uli) c*(c-1)*(c-2)/6
                        + (uli) d*(d-1)*(d-2)*(d-3)/24;
        qnum *= 3;
        lhinfo[qnum]     = d1;
        lhinfo[qnum + 1] = d2;
        lhinfo[qnum + 2] = d3;
} /* writequartetlh */

/************************/

unsigned char readquartet(int a, int b, int c, int d, 
                          unsigned char *quartetinfo)
{
        uli qnum;

	sort4ints(&a,&b,&c,&d);
        qnum = (uli) a
                        + (uli) b*(b-1)/2
                        + (uli) c*(c-1)*(c-2)/6
                        + (uli) d*(d-1)*(d-2)*(d-3)/24;
        if (qnum % 2 == 0) { /* even number */
                /* bits 0 to 3 */
                return (quartetinfo[qnum/2] & (unsigned char) 15);
        } else { /* odd number */
                /* bits 4 to 7 */
                return ((quartetinfo[(qnum-1)/2] & (unsigned char) 240)>>4);
        }
} /* readquartet */

/************************/

void writequartet(int a, int b, int c, int d, unsigned char info,
                  unsigned char *quartetinfo)
{
        uli qnum;

        qnum = (uli) a
                        + (uli) b*(b-1)/2
                        + (uli) c*(c-1)*(c-2)/6
                        + (uli) d*(d-1)*(d-2)*(d-3)/24;
        if (qnum % 2 == 0) { /* even number */
                /* bits 0 to 3 */
                quartetinfo[qnum/2] =
                        ((quartetinfo[qnum/2] & (unsigned char) 240) |
                        (info & (unsigned char) 15));
        } else { /* odd number */
                /* bits 4 to 7 */
                quartetinfo[(qnum-1)/2] =
                        ((quartetinfo[(qnum-1)/2] & (unsigned char) 15) |
                        ((info & (unsigned char) 15)<<4));
        }
} /* writequartet */

/******************************************************************************/

void initqblock(qblocktype *qb)
{
  qb->fname       = malloc(FILENAMELEN * sizeof(char));
  qb->format[0]   = '-';
  qb->format[1]   = '-';
  qb->fstream     = NULL;
  qb->nspec       = 0;
  qb->translat    = NULL;
  qb->names       = NULL;
  qb->quartetinfo = NULL;
  qb->lhinfo      = NULL;
}

/************************/

void cleanupqblock(qblocktype *qb)
{
  int n;

  if (qb->fname != NULL) 
      strcpy(qb->fname,"");
  if (qb->fstream != NULL) 
      fclose(qb->fstream);
  qb->fstream = NULL;

  if (qb->names != NULL) {
      for (n=0 ; n < qb->nspec; n++) 
           free((qb->names)[n]);
      free(qb->names);
  }

  qb->nspec   = 0;

  if (qb->translat != NULL)
      free(qb->translat);
  qb->translat = NULL;
  
  if (qb->quartetinfo != NULL)
      free(qb->quartetinfo);
  qb->quartetinfo = NULL;
  
  if (qb->lhinfo != NULL)
      free(qb->lhinfo);
  qb->lhinfo = NULL;
}

/**********************/

void fprintfmatrix(FILE *of, ulimatrix mat, ulivector vec, int max, int *trueID)
{
     int m,n;

     for (n=0; n<max; n++) {
       fprintf(of, "( ");
       for (m=0; m<max; m++) {
         if (n == m) fprintf(of, "  - ");
         else        fprintf(of, "%3d ", mat[n][m]);
       }
       fprintf(of, ") %3d  '", vec[n]);
       fputid(of, trueID[n]);
       fprintf(of, "'\n");
     }
     fprintf(of, "  ");
     for (m=0; m<max; m++) fprintf(of, "----", vec[m]);
     fprintf(of, "\n  ");
     for (m=0; m<max; m++) fprintf(of, "%3d ", vec[m]);
     fprintf(of, "\n");
}
/******************************************************************************/
/******************************************************************************/
/******************************************************************************/

void dostep(int argc, char *argv[])
{
  qblocktype  qblock1;
  int         n,m;
  int         x,y;
  int         a,b,c,i;
  int        *trueid;
  int         Maxspc;
  int         Maxbrnch; 
  int e;

  /* new algorithm */
  ONEEDGE   *newedge; 
  int       *newedgeofleaf;
  ulimatrix  penaltymat;
  ulivector  penaltyvec;
  ulivector  missingvec;
  int        newnextedge;
  int        newnextleaf;
  uli        newmininfo; 
  int        newminedge; 
  int        newhowmany;

  int       *rightleafs;
  int        rightnum;
  uli        rightsum;
  uli        rightmat;
  int       *leftleafs;
  int        leftnum;
  uli        leftsum;
  uli        leftmat;

  /* old algorithm */
  ONEEDGE *oldedge; 
  int     *oldedgeofleaf;
  int      oldnextedge;
  int      oldnextleaf;
  uli      oldmininfo; 
  int      oldminedge; 
  int      oldhowmany;

  /********************/
  /* read in quartets */
  /********************/
  initqblock(&qblock1);
  strcpy(qblock1.fname, argv[1]);
  freshreadallquarts(&qblock1);

  fprintf(stderr, "XXXXX\n");
  printverboseallquarts(stdout, qblock1);
  fprintf(stderr, "XXXXX\n");

  spcnames = qblock1.names;
  Maxspc   = qblock1.nspec;
  Maxbrnch = 2*Maxspc - 3;

  /*******************************/
  /* variables for new algorithm */
  /*******************************/
  penaltymat = new_ulimatrix(Maxspc, Maxspc);
  penaltyvec = new_ulivector(Maxspc);
  missingvec   = new_ulivector(Maxspc);

  rightnum = 0;
  rightleafs = malloc(Maxspc * sizeof(int));
  rightsum = 0;
  rightmat = 0;

  leftnum= 0;
  leftleafs  = malloc(Maxspc * sizeof(int));
  leftsum = 0;
  leftmat = 0;

  /*******************/
  /* set input order */
  /*******************/
  fprintf(stderr, "set input order\n");
  trueid = malloc(Maxspc * sizeof(int));
  chooser(Maxspc, Maxspc, trueid);
  for (n=0; n<Maxspc; n++) {
     /* trueid[n] = Maxspc - 1 - n; */
     printf ("%d. %d\n", n, trueid[n]);
  }

  /* initialize tree */
  oldinittree(&oldedge, &oldedgeofleaf, Maxspc, Maxbrnch, &oldnextedge, &oldnextleaf);
  newinittree(&newedge, &newedgeofleaf, Maxspc, Maxbrnch, &newnextedge, &newnextleaf);


  for (i = 3; i < Maxspc; i++) {
fprintf(stderr, "--------------------------------------------------\n");

     for (n=0; n<Maxspc; n++) {
       penaltyvec[n] = 0;
       missingvec[n]   = 0;
       for (m=0; m<Maxspc; m++)
         penaltymat[n][m] = 0;
     }

     printf("-");
     for (n=0; n <= i; n++) 
        printf("%3d ", trueid[n]);
     fputid(stdout, trueid[i]);
     printf("\n");

     for (a = 0; a < i - 2; a++) 
        for (b = a + 1; b < i - 1; b++) 
           for (c = b + 1; c < i; c++) {
             checkquartet(a,b,c,i, trueid, qblock1.quartetinfo,&x,&y);

             if (x != y) {
                /* old algoritm */
                oldincrementedgeinfo(x, y, oldedge, oldedgeofleaf);

                /* new algorithm */
                (penaltymat[x][y])++;
                (penaltymat[y][x])++;
                (missingvec[a])++;
                (missingvec[b])++;
                (missingvec[c])++;
                (missingvec[i])++;
#               ifdef VERBOSE
                   printf("   %d, %d, %d, %d (%d,%d) (%lu)\n", 
                          a,b,c,i, x,y, penaltymat[x][y]);
#               endif
             } else {
                fprintf(stderr, "WARNING: missing quartet (%d,%d,%d,%d)\n",a,b,c,i);
             }
     }

     for (n=0; n<i; n++) {
        fprintf(stderr, "taxon %d occurs %d times\n", n, missingvec[n]);
     }
     /* old algoritm */
     fprintf(stderr, "old: ");
     oldminimumedgeinfo(oldedge, oldedgeofleaf, oldnextedge, oldnextleaf, 
                     &oldminedge, &oldmininfo, &oldhowmany);
oldwritetree(stderr, oldedge, oldedgeofleaf, oldnextedge, oldnextleaf, trueid);
     oldaddnextleaf(oldminedge, oldedge, oldedgeofleaf, Maxspc, Maxbrnch, 
                 &oldnextedge, &oldnextleaf);

     /* new algorithm */
     for (n=0; n<i-1; n++) 
       for (m=n+1; m<i; m++) {
         penaltyvec[n] += penaltymat[n][m];
         penaltyvec[m] += penaltymat[n][m];
     }
     newincrementedgeinfo(newedge, newedgeofleaf, penaltymat, penaltyvec, trueid);

     fprintf(stderr, "new: ");
     newminimumedgeinfo(newedge, newedgeofleaf, newnextedge, newnextleaf, 
                     &newminedge, &newmininfo, &newhowmany);
newwritetree(stderr, newedge, newedgeofleaf, newnextedge, newnextleaf, trueid);
     newaddnextleaf(newminedge, newedge, newedgeofleaf, Maxspc, Maxbrnch, 
                 &newnextedge, &newnextleaf);

     printf("\n");
     fprintfmatrix(stdout, penaltymat, penaltyvec, i, trueid);
     printf("\n");



     for (e=0; e<newnextedge; e++) {
        printf(".");
        if (newedge[e].edgeinfo != oldedge[e].edgeinfo)
          printf("%3d. n%d[n%d] != o%d[n%d]\n", e, newedge[e].edgeinfo, e, oldedge[e].edgeinfo, e);
     }
     printf("\n");
writetree(stderr, newedge, newedgeofleaf, newnextedge, newnextleaf, trueid);

     /* reset */
     oldresetedgeinfo(oldedge, oldnextedge);
     newresetedgeinfo(newedge, newnextedge);


  }
  cleanupqblock(&qblock1);
}

/******************************************************************************/

void printusage(char *argv[])
{
      printf("\nUSAGE: %s -h | quartfile \n\n", argv[0]);
      printf("\t-h         - this help text\n");
      printf("\n");
}

/******************************************************************************/
/******************************************************************************/
/******************************************************************************/

int main(int argc, char *argv[])
{
   initrandom(-1);
   /* initrandom(5); */
   if ((argc < 2) || (0 == strcmp(argv[1], "-h"))) {
      printusage(argv);
      exit(0);
   } else {
      dostep(argc,argv);
   } 
   exit(0);
}
